package oracle.spatial.raster;

import oracle.spatial.hadoop.imageprocessor.datatype.ImageBandWritable;

import org.apache.hadoop.io.DoubleWritable;


public class HillShadeUtil {

    /**
     * Hypothetical altitude of the ilumination in degrees above horizontal
     */
    private int altitude = 45;
    
    /**
     * Hypothetical altitude of the ilumination in degrees
     */
    private int azimuth = 315;
    
    /**
     * Coefficient that related DEM reference units to value units
     */
    private double zFactor = 1;
    
    /**
     * Rate of change in X direction , changes for every pixel 
     */
    double dx;
    
    /**
     * Rate of change in Y direction , changes for every pixel 
     */
    double dy;

   

    /**
     * Clears the neighborhood matrix.
     *
     * @param neighborhood  the matrix to clean.
     */
    private void clearNeighborhood(double[] neighborhood) {
        for (int i = 0; i < neighborhood.length; i++) {
            neighborhood[i] = 0;
        }
    }


/**
     * Calculates hillshade
     * @param zenithRadians     zenith value in radians
     * @param azimuthRadians    azimuth value in radians
     * @param slopeRadians      slope value in radians
     * @param aspectRadians     aspect value in radians
     * @return
     */
    private float calculateHillShade(double zenithRadians, double azimuthRadians, double slopeRadians,
                                     double aspectRadians) {
        float hillShade = 0;
        hillShade =
            (float) (255 *
                     ((Math.cos(zenithRadians) * Math.cos(slopeRadians)) +
                      (Math.sin(zenithRadians) * Math.sin(slopeRadians) * Math.cos(azimuthRadians - aspectRadians))));
        return hillShade;
    }

    /**
     * Generates a value for a specific neighboor of the evaluated pixel, depending
     * on its position, if it is beyon tile borders a zero value is set.
     *
     * @param width             tile width.
     * @param data          array with values for pixels in the tile.
     * @param xReference        neighboor requested X position.
     * @param rightColumn       <code>true</code> if neighboor requested is on the
     *                          right column of the matrix.
     * @param lastValue         the las possible position in the bandData array.
     * @param height            tile height.
     * @param yReference        neighboor requested Y position.
     * @param upperRow          <code>true</code> if neighboor requested is on the
     *                          upper row of the matrix.
     * @param isCenterColumn    <code>true</code> if neighboor requested is on the
     *                          center column of the matrix.
     * @param centerRow         <code>true</code> if neighboor requested is on the
     *                          center row of the matrix.
     * @return                  the value for the required neighboor.
     */
    private double calculateNeighborValue(int width, float[] data, int xReference, boolean rightColumn, int lastValue,
                                          int height, int yReference, boolean upperRow, boolean isCenterColumn,
                                          boolean centerRow) {

        double neighboor = 0;

        if (yReference < 0 || yReference > height) {
            return neighboor;
        } else if (!isCenterColumn && !rightColumn && (xReference < 0 || xReference % width == width - 1)) {
            return neighboor;
        } else if (!isCenterColumn && rightColumn && (xReference % width == 0 || xReference > lastValue)) {
            return neighboor;
        } else if (upperRow) {
            if (xReference - width >= 0) {

                neighboor = data[xReference - width];

            } else {
                return neighboor;
            }
        } else if (centerRow) {
            if (xReference < lastValue) {
                neighboor = data[xReference];
            } else {
                return neighboor;
            }
        } else {
            if (xReference + width < lastValue) {
                neighboor = data[xReference + width];

            } else {
                return neighboor;
            }
        }
        return neighboor;
    }

    /**
     * Fills the neighboor matrix for a specified pixel. Matrix is 3 x 3 containing
     * the surrounding pixels of the one being valuated.
     *
     * @param width         tile width.
     * @param bandData      array with values for pixels in the tile.
     * @param lastValue     the las possible position in the bandData array.
     * @param height        tile height.
     * @param position      the position of the pixel to evaluate.
     * @param neighborhood  matrix of neighboors to fill.
     * @param row           the row where evaluated pixel is located.
     */
    private void fillNeighborhoodValue(int width, float[] bandData, int lastValue, int height, int position,
                                       double[] neighborhood, int row) {
        neighborhood[0] =
            calculateNeighborValue(width, bandData, position - 1, false, lastValue, height, row - 1, true, false,
                                   false);
        neighborhood[1] =
            calculateNeighborValue(width, bandData, position, false, lastValue, height, row - 1, true, true, false);
        neighborhood[2] =
            calculateNeighborValue(width, bandData, position + 1, true, lastValue, height, row - 1, true, false, false);
        neighborhood[3] =
            calculateNeighborValue(width, bandData, position - 1, false, lastValue, height, row, false, false, true);
        neighborhood[4] =
            calculateNeighborValue(width, bandData, position, false, lastValue, height, row, false, true, true);
        neighborhood[5] =
            calculateNeighborValue(width, bandData, position + 1, true, lastValue, height, row, false, false, true);
        neighborhood[6] =
            calculateNeighborValue(width, bandData, position - 1, false, lastValue, height, row + 1, false, false,
                                   false);
        neighborhood[7] =
            calculateNeighborValue(width, bandData, position, false, lastValue, height, row + 1, false, true, false);
        neighborhood[8] =
            calculateNeighborValue(width, bandData, position + 1, true, lastValue, height, row + 1, false, false,
                                   false);
    }


/**
     * Calculates the hillshade value for the pixel data received
     * @param lastValue     the las possible position in the bandData array.
     * @param pixelXWidth   X width for this raster
     * @param pixelYWidth   Y width for this raster
     * @param noData        NODATA value for this raster
     * @param image         Raster data type to store hillshade pixels
     * @param newWidth      width of the tile without overlapping pixels
     * @param newHeight     height of the tile without overlapping pixels
     * @param band          Band to process hillshade
     * @param startX        start pixel in X to analyze in the tile
     * @param startY        start pixel in Y to analyze in the tile
     * @param endX          end pixel in X to analyze in the tile
     * @param endY          end pixel in Y to analyze in the tile
     * @param width         tile width.
     * @param height        tile height
     */
   public void hillshade(int lastValue, double pixelXWidth, double pixelYWidth, DoubleWritable noData,
                          ImageBandWritable image, int newWidth, int newHeight, int band, int startX, int startY,
                          int endX, int endY, int width, int height) {
        float noDataValue = 0;

        int zenithDegrees = 90 - altitude;
        double zenithRadians = zenithDegrees * Math.PI / 180;

        int azimuth_math = 360 - azimuth + 90;

        if (azimuth_math > 360) {
            azimuth_math = azimuth_math - 360;
        }

        double azimuthRadians = azimuth_math * Math.PI / 180;

        if (noData != null) {
            noDataValue = (float) noData.get();
            System.out.println("No data value for band " + band + " is " + noDataValue);
            noData = null;
        }

        float[] finalFloatData = new float[newWidth * newHeight];

        float[] bandData = (float[]) image.getBand(band - 1);

        double[] neighborhood = new double[9];
        int position = 0;
        float slope = 0;
        double slopeRadians = 0;
        float hillShade = 0;
        double aspectRadians = 0;


      

        for (int jj = startY; jj < endY; jj++) {

            for (int ii = startX; ii < endX; ii++) {
                slope = noDataValue;
                try {

                    fillNeighborhoodValue(width, bandData, lastValue, height, (width * jj) + ii, neighborhood, jj);


                } catch (Exception e) {
                    System.out.println("Problem while filling neighbor matrix" + e.getMessage());
                }


                for (int i = 0; i < 9; i++) {
                    if (neighborhood[i] == noDataValue) {
                        slope = 0;
                        break;
                    }
                }
                if (slope == noDataValue) {
                    slope = (float) calculateSlope(neighborhood, pixelXWidth, pixelYWidth);
                }

                clearNeighborhood(neighborhood);
                slopeRadians = Math.atan(slope * zFactor);
                aspectRadians = calculateAspectRadians();
                hillShade = calculateHillShade(zenithRadians, azimuthRadians, slopeRadians, aspectRadians);
          
                finalFloatData[position] = hillShade;
                position++;
            }
        }
        
        image.replaceBand(finalFloatData, band - 1);
        finalFloatData = null;
    }


/**
     * Calculates the aspect for active pixel in radians
     * @return aspect in radians for active pixel
     */
    private double calculateAspectRadians() {
        double aspectRadians = 0;
        if (dx != 0) {
            aspectRadians = Math.atan2((double) dy, (double) -dx);
            if (aspectRadians < 0) {
                aspectRadians = (2 * Math.PI) + aspectRadians;
            }

        } else {
            if (dy > 0) {
                aspectRadians = Math.PI / 2;
            } else if (dy < 0) {
                aspectRadians = ((2 * Math.PI) - (Math.PI / 2));
            } else {

            }
        }

        return aspectRadians;
    }


    /**
     * Performs the arithmetic operation to calculate the slope of a pixel, according
     * to the value of its surrounding pixels.
     *
     * @param pixelWidthX   width of mosaic pixel in X.
     * @param pixelWidthY   width of mosaic pixel in Y.
     * @param neighboors    array containing center value to calculate slope and
     *                      its 8 neighboors.
     * @return              slope for center value.
     */
    private double calculateSlope(double[] neighboors, double pixelWidthX, double pixelWidthY) {
        double slope = 0;
        dx =
            ((neighboors[0] + (2 * neighboors[3]) + neighboors[6]) -
             (neighboors[2] + (2 * neighboors[5]) + neighboors[8])) / (8 * pixelWidthX);
        dy =
            ((neighboors[6] + (2 * neighboors[7]) + neighboors[8]) -
             (neighboors[0] + (2 * neighboors[1]) + neighboors[2])) / (8 * pixelWidthY);
        slope = Math.sqrt(Math.pow(dx, 2) + Math.pow(dy, 2));
        return slope;
    }


}
