package hol;

import java.io.IOException;

import oracle.spatial.hadoop.vector.geojson.GeoJsonRecordInfoProvider;
import oracle.spatial.hadoop.vector.geojson.mapred.GeoJsonInputFormat;
import oracle.spatial.hadoop.vector.mapred.input.SpatialFilterInputFormat;
import oracle.spatial.hadoop.vector.util.JobUtils;
import oracle.spatial.hadoop.vector.util.SpatialOperation;
import oracle.spatial.hadoop.vector.util.SpatialOperationConfig;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.NullWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapred.JobClient;
import org.apache.hadoop.mapred.JobConf;
import org.apache.hadoop.mapred.MapReduceBase;
import org.apache.hadoop.mapred.Mapper;
import org.apache.hadoop.mapred.OutputCollector;
import org.apache.hadoop.mapred.Reporter;
import org.apache.hadoop.mapred.TextOutputFormat;
import org.apache.hadoop.util.GenericOptionsParser;
import org.codehaus.jackson.JsonNode;
import org.codehaus.jackson.map.ObjectMapper;


/**
 * Sample job to filter spatially and non spatially. 
 * This example filters the tweets in a query window. 
 * 
 * The job expects the following parameters:
 * 		1)	input data path (tweets.json)
 * 		2) 	output path. HDFS folder with the resulting filtered records.
 * 
 */
public class CustomFiltering{

	public static class Map extends MapReduceBase implements
			Mapper<LongWritable, Text, NullWritable, Text> {

		public void map(LongWritable key, Text value,
				OutputCollector<NullWritable, Text> output, Reporter reporter) throws IOException {
			ObjectMapper jsonMapper = new ObjectMapper();
			JsonNode recordNode = jsonMapper.readTree(value.toString());
			long followersCount = recordNode.get("properties").get("followers_count").getLongValue();
			//additionally filter the tweets with more than 50 followers
			if(followersCount > 50){
				output.collect(NullWritable.get(), value);
			}
		}
	}
	 
	public static void main(String[] init_args) throws Exception{
		Configuration config = new Configuration();
		
		// This step is important as init_args contains ALL the arguments passed to hadoop on the command
		// line (such as -libjars [jar files]).  What's left after .getRemainingArgs is just the 
		// application specific arguments
		String [] args = new GenericOptionsParser(config, init_args).getRemainingArgs();
		
		JobConf conf = new JobConf(config);

		//set input path and format
		JobUtils.setupInputPath(args[0], conf);
		//InputFormat that will filter spatially the data
		conf.setInputFormat(SpatialFilterInputFormat.class);
		 
		//set internal input format
		SpatialFilterInputFormat.setInternalInputFormatClass(conf, GeoJsonInputFormat.class);
		//as no spatial index is used a RecordInfoProvider is needed
		SpatialFilterInputFormat.setRecordInfoProviderClass(conf, GeoJsonRecordInfoProvider.class);
		 
		//set spatial operation used to filter the records
		SpatialOperationConfig spatialOperationConfig = new SpatialOperationConfig(); 
		spatialOperationConfig.setSrid(8307);
		spatialOperationConfig.setTolerance(0.5);
		spatialOperationConfig.setGeodetic(true);
		spatialOperationConfig.setOperation(SpatialOperation.IsInside);
		//set the query window (the query window covers an area in the North of Mexico)
		spatialOperationConfig.setJsonQueryWindow("{\"type\":\"Polygon\", \"coordinates\":[[-106, 25, -106, 30, -104, 30, -104, 25, -106, 25]]}");
		
		spatialOperationConfig.store(conf);

		// output path
		JobUtils.setupOutputPath(args[1], null, conf);

		// output format
		conf.setOutputFormat(TextOutputFormat.class);
		// mapper
		conf.setMapperClass(Map.class);
		conf.setMapOutputKeyClass(NullWritable.class);
		conf.setMapOutputValueClass(Text.class);
		//one reducer will be used to avoid multiple outputs
		conf.setNumReduceTasks(1);
		
		// run job
		conf.setJarByClass(CustomFiltering.class);
		conf.setJobName("CustomFiltering example");
		
		JobClient.runJob(conf);
	}
}
