package hol;

import java.io.IOException;
import java.text.DecimalFormat;

import oracle.spatial.geometry.J3D_Geometry;
import oracle.spatial.geometry.JGeometry;
import oracle.spatial.hadoop.vector.RecordInfo;
import oracle.spatial.hadoop.vector.mapred.input.CompositeInputFormat;
import oracle.spatial.hadoop.vector.mapred.input.FileSplitInputFormat;
import oracle.spatial.hadoop.vector.mapred.input.RecordInfoLoader;
import oracle.spatial.hadoop.vector.shapefile.ShapeFileRecordInfoProvider;
import oracle.spatial.hadoop.vector.shapefile.mapred.ShapeFileInputFormat;
import oracle.spatial.hadoop.vector.util.ConfigParams;
import oracle.spatial.hadoop.vector.util.JobUtils;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.io.NullWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.io.Writable;
import org.apache.hadoop.mapred.FileSplit;
import org.apache.hadoop.mapred.JobClient;
import org.apache.hadoop.mapred.JobConf;
import org.apache.hadoop.mapred.MapReduceBase;
import org.apache.hadoop.mapred.Mapper;
import org.apache.hadoop.mapred.OutputCollector;
import org.apache.hadoop.mapred.Reporter;
import org.apache.hadoop.mapred.TextOutputFormat;
import org.apache.hadoop.util.GenericOptionsParser;


/**
 * Sample job that receives as input an ESRI Shapefile containing polygons. 
 * The return are the records attributes and polygons length.
 * 
 * The job expects the following parameters:
 * 		1)	input data path
 * 		2) 	output path. HDFS folder with the resulting records 
 * 
 */
public class PolygonLength {

	public static class Map extends MapReduceBase implements Mapper<NullWritable, FileSplit, NullWritable, Text>{
		private RecordInfoLoader<Writable, Writable> riLoader = null;

		@Override
		public void configure(JobConf conf) {
			super.configure(conf);
			//create the record info loader
			riLoader = new RecordInfoLoader<Writable, Writable>(conf);
		}
		
		@Override
		public void map(NullWritable key, FileSplit value,
				OutputCollector<NullWritable, Text> output, Reporter reporter)
				throws IOException {
			//set a maximum number of results to process to reduce the time of the job
			int maxNumberOfResults = 10;
			int currentResultNumber = 1;
			//Semi-Major Axis (Equatorial Radius)
			double WGS84_SMAX = 6378137;
			//inverse flattening
			double WGS84_IFLAT = 298.257223563;

			riLoader.startLoading(value, reporter);
			
			while(riLoader.hasNext()){
				if(currentResultNumber > maxNumberOfResults){
					//return since the maximum number of results to process has been reached
					return;
				}
				
				currentResultNumber++;
				
				//get the RecordInfo
				RecordInfo ri = riLoader.next();
				
				if(ri!=null && ri.getGeometry() != null){
					try {
						//the class J3D_Geometry contains the length function. It works with
						//both 3D and 2D geometries.
						J3D_Geometry geometry = JGeometry.make_3dgeom(ri.getGeometry(), 
								false, //doesn't ignore SRID conversion
								ri.getGeometry().getSRID(), //geometries SRID
	                            0 //height that is 0 since in our case the geometry is 2D
	                            );
						
						double length = geometry.length(
								0.05 //tolerance
								, "TRUE" //is_g3d TRUE means that the data are geodetic
								, WGS84_SMAX
								, 1/WGS84_IFLAT
								, 1 //use 1 for the unit-of-measure factor (to return meters)
						);
		
						//collect the information to print as result
						StringBuffer sb = new StringBuffer();
						
						sb.append("----------------------------------");
						sb.append("\npostal code reference : " +ri.getField("postcode"));
						sb.append("\nCountry abbreviation : " +ri.getField("iso_ctry"));
						sb.append("\nCountry : " +ri.getField("admin1"));
						sb.append("\nState : " +ri.getField("admin2"));
						sb.append("\nCounty : " +ri.getField("admin3"));
						sb.append("\nCity : " +ri.getField("admin4"));
						sb.append("\nState abbreviation : " +ri.getField("state"));

						sb.append("\nlength : " +new DecimalFormat(".##").format(length));
						sb.append("\n----------------------------------");
						//collect the result
						output.collect(NullWritable.get(), new Text(sb.toString()));

					} catch (Exception e) {
						e.printStackTrace();
					}
				}
			}
		}
	}
	
	public static void main(String[] init_args) throws Exception{
		Configuration config = new Configuration();
		
		// This step is important as init_args contains ALL the arguments passed to hadoop on the command
		// line (such as -libjars [jar files]).  What's left after .getRemainingArgs is just the 
		// application specific arguments
		String [] args = new GenericOptionsParser(config, init_args).getRemainingArgs();
		
		JobConf conf = new JobConf(config);
		conf.setJarByClass(PolygonLength.class);
		conf.setJobName("PolygonLength example");
		//one reducer will be used
		conf.setNumReduceTasks(1);
		
		//Setting add extra fields to true means that all the fields available in the Shapefile will
		//be available when the records will be process in the mapper.
		ShapeFileRecordInfoProvider.setAddExtraFields(conf, true);

		//set input and output paths in HDFS
		JobUtils.setupInputPath(args[0], conf);
		JobUtils.setupOutputPath(args[1], null, conf);
		
		//set the mapper class
		conf.setMapperClass(Map.class);
		
		conf.setMapOutputKeyClass(NullWritable.class);
		conf.setMapOutputValueClass(Text.class);
		conf.setOutputKeyClass(NullWritable.class);
		conf.setOutputValueClass(RecordInfo.class);
		//A CompositeInputFormat that only provides the source FileSplits as its values so mapreduce 
		//components can read the data directly using its internal input format. 
		conf.setInputFormat(FileSplitInputFormat.class);
		
		//set SRID of the geometries
		conf.setInt(ConfigParams.SRID, 4326);
		
		//set the input format class
		CompositeInputFormat.setInternalInputFormatClass(conf, ShapeFileInputFormat.class);
		//set the record info provider class
		CompositeInputFormat.setRecordInfoProviderClass(conf, ShapeFileRecordInfoProvider.class);
		//set the output format class
		conf.setOutputFormat(TextOutputFormat.class);
		//run the job
		JobClient.runJob(conf);
	}
}
